<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:qbe="http://www.in2p3.fr/symod">
    <xsl:output method="html"/>
    <xsl:variable name="BASE">/symod</xsl:variable>

    <xsl:template match="/">
        <html>
            <head>
                <style type="text/css">
table,tr,td,th {
    font-family:tahoma, arial, verdana, sans-serif;
    border-collapse: collapse;
    border-spacing: 0;
    border : solid 0px #BBD;
    margin:0;padding:4px;
    font-size:12px;
}
table { border-bottom : solid 1px #BBD;}
td,th { text-align:center; border-top : solid 1px #BBD;border-left : solid 1px #BBD;border-right : solid 1px #BBD;}
th {
    text-align:left
    min-width:80px;
    line-height:15px;
    background-color:lightgrey;
    background-image:-webkit-gradient(linear, 50% 0%, 50% 100%, color-stop(0%, #f9f9f9), color-stop(100%, #e3e4e6));
    background-image:-webkit-linear-gradient(top, #f9f9f9,#e3e4e6);
    background-image:-moz-linear-gradient(top, #f9f9f9,#e3e4e6);
    background-image:-o-linear-gradient(top, #f9f9f9,#e3e4e6);
    background-image:-ms-linear-gradient(top, #f9f9f9,#e3e4e6);
    background-image:linear-gradient(top, #f9f9f9,#e3e4e6);
    font-size:11px;
}
caption {
    font-weight: bold;
}
                </style>
            </head>
            <body>
                <xsl:apply-templates select="response"/>
            </body>
        </html>
    </xsl:template>
    <xsl:template match="response">
        <h3>
            <xsl:value-of select="qbe:query/qbe:description/text()"/>
        </h3>
        <h4>Nombre d'éléments :
            <xsl:value-of select="count(row)"/>
        </h4>
        <table>
            <thead>
                <xsl:for-each select="header">
                    <tr>
                        <xsl:for-each select="col">
                            <th><xsl:value-of select="text()"/></th>
                        </xsl:for-each>
                    </tr>
                </xsl:for-each>
            </thead>
            <tbody>
                <xsl:for-each select="row">
                    <tr>
                        <xsl:for-each select="col">
                            <td>
                                <xsl:variable name="current" select="position()"/>
                                <xsl:choose>
                                    <xsl:when test="/response/header/col[position()=$current]/@entity">
                                        <!-- absolute path -->
                                        <xsl:variable name="id_url_encode"><xsl:call-template name="url-encode">
                                            <xsl:with-param name="str" select="text()"/>
                                        </xsl:call-template></xsl:variable>
                                        <a href="{$BASE}/entity/{/response/header/col[position()=$current]/@entity}/{$id_url_encode}">
                                            <xsl:value-of select="text()"/>
                                        </a>
                                    </xsl:when>
                                    <xsl:otherwise>
                                        <xsl:value-of select="text()"/>
                                    </xsl:otherwise>
                                </xsl:choose>
                            </td>
                        </xsl:for-each>
                    </tr>
                </xsl:for-each>
            </tbody>
        </table>
    </xsl:template>
    <xsl:variable name="ascii"><![CDATA[ !"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~]]></xsl:variable>
    <xsl:variable name="latin1"><![CDATA[ ¡¢£¤¥¦§¨©ª«¬­®¯°±²³´µ¶·¸¹º»¼½¾¿ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖ×ØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõö÷øùúûüýþÿ]]></xsl:variable>
    <!-- Characters that usually don't need to be escaped -->
    <xsl:variable name="safe"><![CDATA[]!'()*-.0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz~]]></xsl:variable>
    <xsl:variable name="hex">0123456789ABCDEF</xsl:variable>
    <xsl:template name="url-encode">
        <xsl:param name="str"/>
        <xsl:if test="$str">
            <xsl:variable name="first-char" select="substring($str,1,1)"/>
            <xsl:choose>
                <xsl:when test="contains($safe,$first-char)">
                    <xsl:value-of select="$first-char"/>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:variable name="codepoint">
                        <xsl:choose>
                            <xsl:when test="contains($ascii,$first-char)">
                                <xsl:value-of select="string-length(substring-before($ascii,$first-char)) + 32"/>
                            </xsl:when>
                            <xsl:when test="contains($latin1,$first-char)">
                                <xsl:value-of select="string-length(substring-before($latin1,$first-char)) + 160"/>
                            </xsl:when>
                            <xsl:otherwise>
                                <xsl:message terminate="no">
                                    Warning: string contains a character that is out of range! Substituting "?".
                                </xsl:message>
                                <xsl:text>63</xsl:text>
                            </xsl:otherwise>
                        </xsl:choose>
                    </xsl:variable>
                    <xsl:variable name="hex-digit1" select="substring($hex,floor($codepoint div 16) + 1,1)"/>
                    <xsl:variable name="hex-digit2" select="substring($hex,$codepoint mod 16 + 1,1)"/>
                    <xsl:value-of select="concat('%',$hex-digit1,$hex-digit2)"/>
                </xsl:otherwise>
            </xsl:choose>
            <xsl:if test="string-length($str) > 1">
                <xsl:call-template name="url-encode">
                    <xsl:with-param name="str" select="substring($str,2)"/>
                </xsl:call-template>
            </xsl:if>
        </xsl:if>
    </xsl:template>

</xsl:stylesheet>