<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:ent="http://www.in2p3.fr/symod/schema"
                xmlns:symod="http://www.in2p3.fr/symod" exclude-result-prefixes="ent symod"
                xmlns="http://www.in2p3.fr/symod">
    <xsl:output method="xml" indent="yes"/>
    <xsl:variable name="entities" select="document('entities.xml')/*"/>
    <xsl:variable name="query" select="/symod:query"/>

    <xsl:template match="/symod:query">
        <!-- check -->
        <xsl:if test="@_parsedCached and @_parsedCached!=0 and @_parsedCached!=symod:parsedTime/@resolution">
            <xsl:message terminate="yes">Cache resolution must be equal to time resolution</xsl:message>
        </xsl:if>
        <xsl:if test="count(.//symod:print) = 0">
            <xsl:message terminate="yes">Please set at least one column to be printed</xsl:message>
        </xsl:if>
        <xsl:if test="symod:entity/symod:value[@formula and not(symod:print) and not(@name)]">
            <xsl:message terminate="yes">Unused @formula : you should print it, name it or remove it</xsl:message>
        </xsl:if>
        <xsl:if test="@distinct='true' and descendant::*/@order">
            <xsl:message terminate="yes">Attribute "distinct" is incompatible with attribute "order"</xsl:message>
        </xsl:if>
        <xsl:for-each select="symod:entity/symod:value | symod:entity/symod:entity_value">
            <xsl:if test="@formula[contains(.,'AVG') or contains(.,'COUNT') or contains(.,'MAX') or contains(.,'MIN') or contains(.,'SUM')]
                   and @constraint[contains(.,'AVG') or contains(.,'COUNT') or contains(.,'MAX') or contains(.,'MIN') or contains(.,'SUM')]">
                <xsl:message terminate="yes">Using two aggregates on the same value is not allowed</xsl:message>
            </xsl:if>
            <xsl:if test="@join and not(@join=parent::symod:entity/preceding-sibling::symod:entity/symod:value/@join
                                     or @join=parent::symod:entity/following-sibling::symod:entity/symod:value/@join)">
                <xsl:message terminate="yes">Orphelin join is not allowed: "<xsl:value-of select="@join"/>"</xsl:message>
            </xsl:if>
        </xsl:for-each>

        <!-- add schema info -->
        <query database="{$entities/@dbname}">
            <xsl:copy-of select="@*"/>
            <xsl:if test="@_parsedCached = '0'">
                <xsl:attribute name="cached"><xsl:value-of select="symod:parsedTime/@resolution"/></xsl:attribute>
            </xsl:if>
            <!-- copy entities with no alternative time join at all, or with at list one attribute/measure with no alternative time join -->
            <xsl:apply-templates select="symod:description | symod:parsedTime
                | symod:entity[not(symod:value[@joinTime='false' or symod:since|symod:until|symod:period])]
                | symod:entity[symod:value[(@TYPE!='Id' and not(@joinTime='false' or symod:since|symod:until|symod:period))]]"/>

            <!-- duplicate entities with an alternative time join -->
            <xsl:for-each select="symod:entity/symod:value[@joinTime='false' or symod:since|symod:until|symod:period]">
                <entity>
                    <xsl:copy-of select="ancestor::symod:entity/@*[not(local-name()='constraint')]"/>
                    <value TYPE="Id" join="join_{ancestor::symod:entity/@TYPE}"/>
                    <xsl:apply-templates select="."/>
                </entity>
            </xsl:for-each>

            <xsl:apply-templates select="symod:global_value"/>
        </query>
    </xsl:template>

    <xsl:template match="symod:entity">
        <!-- find corresponding schema -->
        <xsl:variable name="entity" select="@TYPE"/>
        <xsl:variable name="schema" select="$entities/ent:entity[@id=$entity]"/>
        <xsl:if test="not($schema)">
            <xsl:message terminate="yes">Unknown entity type: <xsl:value-of select="$entity"/></xsl:message>
        </xsl:if>
        <xsl:for-each select="symod:value[@TYPE!='Id'][not(@TYPE=$schema/*/ent:value/@id)]">
            <xsl:message terminate="yes">Unknown value type: <xsl:value-of select="$entity"/>.<xsl:value-of select="@TYPE"/></xsl:message>
        </xsl:for-each>

        <entity>
            <xsl:copy-of select="@*"/>
            <xsl:apply-templates select="symod:value[@TYPE!='Id'][@TYPE=$schema/ent:A/ent:value/@id][not(@joinTime='false' or symod:since|symod:until|symod:period)]"/>
            <xsl:apply-templates select="symod:value[@TYPE!='Id'][@TYPE=$schema/ent:M/ent:value/@id][not(@joinTime='false' or symod:since|symod:until|symod:period)]"/>
            <xsl:apply-templates select="symod:value[@TYPE='Id']"/>
            <!-- add pre-processed value 'Id' if it is missing -->
            <xsl:if test="not(symod:value[@TYPE='Id'])">
                <value _shortTYPE="Id" TYPE="Id" join="join_{@TYPE}"/>
            </xsl:if>

            <xsl:apply-templates select="symod:entity_value"/>
        </entity>
    </xsl:template>

    <xsl:template match="symod:value[@TYPE='Id']">
        <value _shortTYPE="Id"><xsl:copy-of select="@*"/>
            <!-- join to occurences of this entity with an alternative time join -->
            <xsl:if test="not(@join) and parent::symod:entity/symod:value[@joinTime='false' or symod:since|symod:until|symod:period]">
                <xsl:attribute name="join">join_<xsl:value-of select="ancestor::symod:entity/@TYPE"/></xsl:attribute>
            </xsl:if>
            <xsl:apply-templates/>
        </value>
    </xsl:template>
    <xsl:template match="symod:value[@TYPE!='Id']">
        <!-- find corresponding schema -->
        <xsl:variable name="entity" select="ancestor::symod:entity/@TYPE"/>
        <xsl:variable name="value" select="@TYPE"/>
        <xsl:variable name="schema" select="$entities/ent:entity[@id=$entity]/*/ent:value[@id=$value]"/>
        <xsl:if test="not($schema)">
            <xsl:message terminate="yes">Unknown value type: <xsl:value-of select="$value"/></xsl:message>
        </xsl:if>
        <xsl:if test="@link and not(@link=$schema/@foreignEntity)">
            <xsl:message terminate="yes">Value <xsl:value-of select="$value"/> is not linked to <xsl:value-of select="@link"/></xsl:message>
        </xsl:if>

        <!-- add schema info -->
        <value _shortTYPE="{$schema/@shortid}" _isDate="{$schema/@field='VALUE_DATE'}" _table="{name($schema/parent::*)}">
            <xsl:apply-templates select="@* | * | text()"/>
        </value>
    </xsl:template>

    <xsl:template match="symod:print">
        <print>
            <xsl:copy-of select="@*"/>
            <xsl:if test="not(@position)">
                <xsl:attribute name="position">99</xsl:attribute>
            </xsl:if>
            <xsl:if test="not(@as)">
                <xsl:choose>
                    <xsl:when test="parent::symod:*/@name">
                        <xsl:attribute name="as">
                            <xsl:value-of select="parent::symod:*/@name"/>
                        </xsl:attribute>
                    </xsl:when>
                    <xsl:when test="parent::symod:parsedTime">
                        <xsl:attribute name="as">Time</xsl:attribute>
                    </xsl:when>
                </xsl:choose>
            </xsl:if>
            <xsl:apply-templates/>
        </print>
    </xsl:template>

    <xsl:template match="/symod:query[@_parsedCached and @_parsedCached!=0 and symod:parsedTime/@resolution]/symod:parsedTime/@since">
        <xsl:attribute name="since"><xsl:value-of select="parent::symod:parsedTime/@until"/> - <xsl:value-of select="parent::symod:parsedTime/@resolution"/>/1440</xsl:attribute>
        <xsl:attribute name="oldSince"><xsl:value-of select="."/></xsl:attribute>
    </xsl:template>
    <xsl:template match="/symod:query[@_parsedCached and @_parsedCached!=0 and symod:parsedTime/@resolution]/symod:parsedTime/@singlePlot">
        <xsl:attribute name="singlePlot">true</xsl:attribute>
    </xsl:template>

    <!-- default templates -->
    <xsl:template match="*">
        <xsl:copy>
            <xsl:apply-templates select="@*|*|text()"/>
        </xsl:copy>
    </xsl:template>
    <xsl:template match="@* | text()">
        <xsl:copy-of select="."/>
    </xsl:template>
</xsl:stylesheet>
