<?xml version="1.0"?>
<xsl:stylesheet version="1.0"
                xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:symod="http://www.in2p3.fr/symod" exclude-result-prefixes="symod"
                xmlns="http://symod.in2p3.fr/sql">
    <xsl:output method="xml" indent="yes"/>

    <xsl:template match="/symod:query">
        <query>
            <xsl:copy-of select="@database | @plot | @singlePlot | @optimize | @oldSince | @since | @until | @resolution"/>
            <xsl:if test="@_parsedCached">
                <xsl:attribute name="cacheResolution"><xsl:value-of select="@_parsedCached"/></xsl:attribute>
            </xsl:if>

            <select>
                <xsl:copy-of select="@distinct"/>
                <xsl:for-each select="//symod:print[
                            parent::symod:parsedTime
                            or (
                                (not(/symod:query/@union='true' and ancestor::symod:query[preceding-sibling::symod:query]))
                                and
                                not(//symod:*[self::symod:value or self::symod:entity_value][@pivot='true'])
                               )
                        ]">
                    <xsl:sort select="@position" data-type="number" order="ascending"/>
                    <xsl:variable name="as" select="@as"/>
                    <xsl:if test="not(preceding::symod:print[@as=$as])">
                        <column>
                            <xsl:copy-of select="@as | @format | parent::symod:value/@limit"/>
                            <xsl:apply-templates select="parent::symod:*" mode="EXTERNAL"/>
                        </column>
                    </xsl:if>
                </xsl:for-each>
                <xsl:if test="//symod:*[self::symod:value or self::symod:entity_value][@pivot='true']">
                    <column as="pivot_description">
                        <_>'<xsl:text/>
                        <xsl:for-each select="//symod:*[self::symod:value or self::symod:entity_value][@pivot='true']">
                            <xsl:sort select="symod:print/@position" data-type="number" order="ascending"/>
                            <xsl:apply-templates select="." mode="NAME"/>
                            <xsl:if test="position()!=last()">, </xsl:if>
                        </xsl:for-each>
                        <xsl:text/>'</_>
                    </column>
                    <column>
                        <xsl:attribute name="as">
                            <xsl:for-each select="//symod:*[self::symod:value or self::symod:entity_value][@pivot='true']">
                                <xsl:sort select="symod:print/@position" data-type="number" order="ascending"/>
                                <xsl:choose>
                                    <xsl:when test="symod:print/@as"><xsl:value-of select="symod:print/@as"/></xsl:when>
                                    <xsl:otherwise>
                                        <xsl:apply-templates select="." mode="NAME"/>
                                    </xsl:otherwise>
                                </xsl:choose>
                                <xsl:if test="position()!=last()">_</xsl:if>
                            </xsl:for-each>
                        </xsl:attribute>
                        <field prefix="r1">
                            <xsl:variable name="pivot_column_name">
                                <xsl:for-each select="//symod:*[self::symod:value or self::symod:entity_value][@pivot='true']">
                                    <xsl:sort select="symod:print/@position" data-type="number" order="ascending"/>
                                    <xsl:apply-templates select="." mode="NAME"/>
                                    <xsl:if test="position()!=last()">_</xsl:if>
                                </xsl:for-each>
                            </xsl:variable>
                            <xsl:attribute name="name">
                                <xsl:value-of select="substring($pivot_column_name, 0, 27)"/>
                                <xsl:text>_XML</xsl:text>
                            </xsl:attribute>
                        </field>
                    </column>
                </xsl:if>
            </select>

            <fromQueries>
                <xsl:attribute name="relation">
                    <xsl:choose>
                        <xsl:when test="@union='true'">UNION ALL</xsl:when>
                        <xsl:when test="symod:parsedTime or
                            (count(symod:query)>1 and
                             symod:query[1]/descendant::symod:value[symod:print or @name][descendant::symod:*/@aggregate] and
                             symod:query[1]/descendant::symod:value[symod:print or @name][not(descendant::symod:*/@aggregate)]
                            )">LEFT OUTER JOIN</xsl:when>
                        <xsl:otherwise>,</xsl:otherwise>
                    </xsl:choose>
                </xsl:attribute>

                <xsl:choose>
                    <!-- ignore if @union is true -->
                    <xsl:when test="@union='true'"/>
                    <!-- time + value references -->
                    <xsl:when test="symod:parsedTime[@resolution]
                                and count(/symod:query/symod:query)>1 and count(descendant::symod:*/@aggregate)>0
                                and count(descendant::symod:value[symod:print or @name][not(descendant::symod:*/@aggregate)])>0">
                        <query declarePrefix="ref">
                            <select/>
                            <fromQueries relation=",">
                                <queryTimestampRef>
                                    <xsl:copy-of select="@resolution | @since | @until"/>
                                    <xsl:apply-templates select="symod:parsedTime" mode="FIELD_TS"/>
                                </queryTimestampRef>
                                <queryUnion>
                                    <xsl:for-each select="/symod:query/symod:query">
                                        <queryRef query="{@name}">
                                            <xsl:apply-templates select="descendant::symod:value[symod:print or @name][not(descendant::symod:*/@aggregate)]"
                                                                 mode="FIELD_TS"/>
                                        </queryRef>
                                    </xsl:for-each>
                                </queryUnion>
                            </fromQueries>
                        </query>
                    </xsl:when>
                    <!-- time reference only -->
                    <xsl:when test="symod:parsedTime[@resolution]">
                        <queryTimestampRef declarePrefix="ref">
                            <xsl:copy-of select="@resolution | @since | @until"/>
                            <xsl:apply-templates select="symod:parsedTime" mode="FIELD_TS"/>
                        </queryTimestampRef>
                    </xsl:when>
                    <!-- value reference only -->
                    <xsl:when test="count(/symod:query/symod:query)>1 and count(descendant::symod:*/@aggregate)>0
                                and count(descendant::symod:value[symod:print or @name][not(descendant::symod:*/@aggregate)])>0">
                        <queryUnion declarePrefix="ref">
                            <xsl:for-each select="/symod:query/symod:query">
                                <queryRef query="{@name}">
                                    <xsl:apply-templates select="descendant::symod:value[symod:print or @name][not(descendant::symod:*/@aggregate)]"
                                                         mode="FIELD_TS"/>
                                </queryRef>
                            </xsl:for-each>
                        </queryUnion>
                    </xsl:when>
                </xsl:choose>

                <xsl:apply-templates select="symod:query"/>
            </fromQueries>

            <where>
                <and>
                    <xsl:for-each select="symod:global_value/symod:constraint">
                        <multi_constraint>
                            <xsl:copy-of select="@*"/>
                            <xsl:value-of select="parent::symod:global_value/@formula"/>
                            <xsl:text> </xsl:text>
                            <xsl:value-of select="text()"/>
                        </multi_constraint>
                    </xsl:for-each>
                </and>
            </where>

            <order>
                <xsl:for-each select="//symod:*[@order][
                            self::symod:parsedTime
                            or (
                                (not(/symod:query/@union='true' and ancestor::symod:query[preceding-sibling::symod:query]))
                                and
                                not(//symod:*[self::symod:value or self::symod:entity_value][@pivot='true'])
                               )
                        ]">
                    <xsl:sort select="@order" data-type="number" order="ascending"/>
                    <by>
                        <xsl:copy-of select="@descending"/>
                        <xsl:apply-templates select="self::symod:*" mode="EXTERNAL"/>
                    </by>
                </xsl:for-each>
            </order>
        </query>
    </xsl:template>

    <xsl:template match="symod:parsedTime" mode="EXTERNAL">
        <xsl:if test="(@resolution or @order) and not(/symod:query/symod:query[1]/descendant::symod:entity[symod:value[@_table='M' or @_isDate='true'][not(@joinTime='false')]])">
            <xsl:message terminate="yes">Can not join query with time.</xsl:message>
        </xsl:if>
        <xsl:choose>
            <xsl:when test="/symod:query/@union='true' or not(/symod:query/@resolution)">
                <field prefix="r1">
                    <xsl:attribute name="name">ts_<xsl:apply-templates select="/symod:query/symod:query[1]/descendant::symod:entity
                        [symod:value[@_table='M' or @_isDate='true']][1]" mode="ENTITY_ID"/></xsl:attribute>
                </field>
            </xsl:when>
            <xsl:otherwise>
                <xsl:apply-templates select="." mode="FIELD_TS"/>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>
    <xsl:template match="symod:value" mode="EXTERNAL">
        <field prefix="{ancestor::symod:query/@name}">
            <xsl:attribute name="name">
                <xsl:apply-templates select="." mode="NAME"/>
            </xsl:attribute>
        </field>
    </xsl:template>
    <xsl:template match="symod:entity_value" mode="EXTERNAL">
        <field prefix="{ancestor::symod:query/@name}" name="{@name}"/>
    </xsl:template>
    <xsl:template match="symod:global_value" mode="EXTERNAL">
        <_><xsl:value-of select="@formula"/></_>
    </xsl:template>

    <xsl:template match="symod:query">
        <query>
            <xsl:if test="not(/symod:query/@union='true' and preceding-sibling::symod:query)">
                <xsl:attribute name="declarePrefix"><xsl:value-of select="@name"/></xsl:attribute>
            </xsl:if>
            <xsl:if test="not(/symod:query/@union='true')
                                  and /symod:query/symod:parsedTime/@resolution">
                <!-- uniquement la 1ere de toutes les measures/value_dates (qui sont jointes avec le temps de reference) -->
                <xsl:for-each select="descendant::symod:entity[symod:value[(@_table='M' and not(preceding-sibling::symod:value/@_table='M')) or @_isDate='true'][not(@joinTime='false')][not(@_parsedSince|@_parsedUntil)]][1]
                                                              /symod:value[(@_table='M' and not(preceding-sibling::symod:value/@_table='M')) or @_isDate='true'][not(@joinTime='false')][not(@_parsedSince|@_parsedUntil)][1]">
                    <outerJoin>
                        <xsl:attribute name="on">ts_<xsl:apply-templates select="parent::symod:entity" mode="ENTITY_ID"/>
                            <xsl:if test="@_isDate='true'">_<xsl:value-of select="@_shortTYPE"/></xsl:if>
                        </xsl:attribute>
                        <xsl:apply-templates select="/symod:query/symod:parsedTime" mode="FIELD_TS"/>
                    </outerJoin>
                </xsl:for-each>
            </xsl:if>
            <xsl:if test="not(/symod:query/@union='true')
                                  and count(/symod:query/symod:query)>1
                                  and descendant::symod:value[symod:print or @name][descendant::symod:*/@aggregate]">
                <xsl:for-each select="descendant::symod:value[symod:print or @name][not(descendant::symod:*/@aggregate)]">
                    <outerJoin>
                        <xsl:attribute name="on">
                            <xsl:apply-templates select="." mode="NAME"/>
                        </xsl:attribute>
                        <field prefix="ref">
                            <xsl:attribute name="name">
                                <!-- get reference name -->
                                <xsl:apply-templates mode="NAME"
select="/symod:query/symod:query[1]/descendant::symod:entity[@TYPE=current()/parent::symod:entity/@TYPE]/symod:value[@TYPE=current()/@TYPE]"/>
                            </xsl:attribute>
                        </field>
                    </outerJoin>
                </xsl:for-each>
            </xsl:if>

            <xsl:choose>
                <xsl:when test="descendant::symod:*[self::symod:value or self::symod:entity_value][@pivot='true']">
                    <select/>
                    <fromQueries>
                        <query>
                            <xsl:call-template name="QUERY"/>
                        </query>
                    </fromQueries>
                    <pivot>
                        <xsl:for-each select="descendant::symod:*[self::symod:value or self::symod:entity_value][not(@pivot='true') and symod:print]">
                            <xsl:sort select="symod:print/@position" data-type="number" order="ascending"/>
                            <column><xsl:copy-of select="symod:print/@as"/>
                                <_>AVG("<xsl:text/>
                                <xsl:apply-templates select="." mode="NAME"/>
                                <xsl:text/>")</_>
                            </column>
                        </xsl:for-each>
                        <xsl:for-each select="descendant::symod:*[self::symod:value or self::symod:entity_value][@pivot='true']">
                            <xsl:sort select="symod:print/@position" data-type="number" order="ascending"/>
                            <for>
                                <xsl:attribute name="column">
                                    <xsl:apply-templates select="." mode="NAME"/>
                                </xsl:attribute>
                            </for>
                        </xsl:for-each>
                    </pivot>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:call-template name="QUERY"/>
                </xsl:otherwise>
            </xsl:choose>
        </query>
    </xsl:template>
    <xsl:template name="QUERY">
            <select>
                <xsl:if test="count(descendant::symod:*[self::symod:value or self::symod:entity_value][@pivot='true'])>0">
                    <xsl:variable name="current" select="."/>
                    <xsl:attribute name="pivot">
                        <xsl:for-each select="descendant::symod:*[self::symod:value or self::symod:entity_value][@pivot='true' and symod:print]">
                            <xsl:sort select="symod:print/@position" data-type="number"/>
                            <xsl:apply-templates select="." mode="VALUE_ID"/>
                            <xsl:if test="position()!=last()">_</xsl:if>
                        </xsl:for-each>
                    </xsl:attribute>
                    <xsl:attribute name="as">
                        <xsl:for-each select="descendant::symod:*[self::symod:value or self::symod:entity_value][@pivot='true' and symod:print]">
                            <xsl:sort select="symod:print/@position" data-type="number"/>
                            <xsl:choose>
                                <xsl:when test="@as"><xsl:value-of select="@as"/></xsl:when>
                                <xsl:otherwise><xsl:apply-templates select="." mode="VALUE_ID"/></xsl:otherwise>
                            </xsl:choose>
                            <xsl:if test="position()!=last()">_</xsl:if>
                        </xsl:for-each>
                    </xsl:attribute>
                    <xsl:for-each select="descendant::symod:*[self::symod:value or self::symod:entity_value][@pivot='true']/descendant::symod:print/@position
                        [. &lt;= $current/descendant::symod:*[self::symod:value or self::symod:entity_value][@pivot='true']/descendant::symod:print/@position]">
                        <xsl:attribute name="position"><xsl:value-of select="."/></xsl:attribute>
                    </xsl:for-each>
                </xsl:if>

                <xsl:if test="/symod:query/symod:parsedTime[symod:print or @name or @resolution or @order]">
                    <!-- uniquement la 1ere de toutes les measures/value_dates -->
                    <xsl:for-each select="descendant::symod:entity[symod:value[(@_table='M' and not(preceding-sibling::symod:value/@_table='M')) or @_isDate='true']][1]
                                                                  /symod:value[(@_table='M' and not(preceding-sibling::symod:value/@_table='M')) or @_isDate='true'][1]">
                        <column>
                            <xsl:attribute name="as">ts_<xsl:apply-templates select="parent::symod:entity" mode="ENTITY_ID"/>
                                <xsl:if test="@_isDate='true'">_<xsl:value-of select="@_shortTYPE"/></xsl:if>
                            </xsl:attribute>
                            <xsl:apply-templates select="." mode="FIELD_TS_ROUNDED"/>
                        </column>
                    </xsl:for-each>
                </xsl:if>

                <xsl:apply-templates select="descendant::symod:print"/>

                <!-- unprinted but named values -->
                <xsl:for-each select="descendant::symod:value[not(symod:print) and @name]">
                    <column as="{@name}">
                        <field name="{@TYPE}">
                            <xsl:attribute name="prefix"><xsl:apply-templates select="." mode="VALUE_ID"/></xsl:attribute>
                            <xsl:apply-templates select="symod:parsedFormula"/>
                        </field>
                    </column>
                </xsl:for-each>
            </select>

            <!-- add joins -->
            <xsl:apply-templates select="symod:entity" mode="JOIN_ENTITY"/>

            <xsl:choose>
                <!-- aggregat -->
                <xsl:when test="count(descendant::symod:*/@aggregate) > 0">
                    <where>
                        <and>
                            <xsl:call-template name="TIME_CONSTRAINTS"/>
                        </and>
                        <and>
                            <xsl:call-template name="NULL_CONSTRAINTS"/>
                        </and>
                        <and>
                            <xsl:apply-templates select="descendant::symod:value[not(symod:print) or descendant::symod:*/@aggregate]
                                                         /symod:parsedConstraint[not(descendant::symod:*/@aggregate)]
                                                       | descendant::symod:entity/@constraint"/>
                        </and>
                    </where>
                    <group>
                        <!-- group by time -->
                        <xsl:if test="/symod:query/symod:parsedTime/@resolution">
                            <xsl:for-each select="descendant::symod:entity[symod:value[(@_table='M' and not(preceding-sibling::symod:value/@_table='M')) or @_isDate='true']][1]
                                                                          /symod:value[(@_table='M' and not(preceding-sibling::symod:value/@_table='M')) or @_isDate='true'][1]">
                                <by>
                                    <xsl:apply-templates select="." mode="FIELD_TS_ROUNDED"/>
                                </by>
                            </xsl:for-each>
                        </xsl:if>
                        <!-- group by selected values -->
                        <xsl:for-each select="descendant::symod:value[symod:print or @name]
                                                                     [not(symod:*/@aggregate)]">
                            <by>
                                <field name="{@TYPE}">
                                    <xsl:attribute name="prefix"><xsl:apply-templates select="." mode="VALUE_ID"/></xsl:attribute>
                                    <xsl:apply-templates select="symod:parsedFormula"/>
                                </field>
                            </by>
                        </xsl:for-each>
                        <!-- group by selected entity_values -->
                        <xsl:for-each select="descendant::symod:entity_value[symod:print or @name]
                                                                            [not(symod:*/@aggregate)]">
                            <by>
                                <formula><xsl:value-of select="symod:parsedFormula/@formula"/></formula>
                            </by>
                        </xsl:for-each>
                    </group>
                    <having>
                        <and>
                            <xsl:apply-templates
                            select="descendant::symod:parsedConstraint[descendant::symod:*/@aggregate or
                                    parent::symod:value[symod:print and not(descendant::symod:*/@aggregate)]]"/>
                        </and>
                    </having>
                </xsl:when>

                <!-- no aggregat -->
                <xsl:otherwise>
                    <where>
                        <and>
                            <xsl:call-template name="TIME_CONSTRAINTS"/>
                        </and>
                        <and>
                            <xsl:call-template name="NULL_CONSTRAINTS"/>
                        </and>
                        <and>
                            <xsl:apply-templates select="descendant::symod:parsedConstraint
                                                       | descendant::symod:entity/@constraint"/>
                        </and>
                    </where>
                </xsl:otherwise>
            </xsl:choose>
    </xsl:template>

    <xsl:template name="TIME_CONSTRAINTS">
        <xsl:for-each select="descendant::symod:entity">
            <xsl:for-each select="symod:value[not(@joinTime='false')][@_table='M'][1] |
                                  symod:value[not(@joinTime='false')][@_isDate='true']">
                <xsl:if test="/symod:query/symod:parsedTime/@since | @_parsedSince">
                    <constraint operator="&gt;=">
                        <xsl:apply-templates select="." mode="FIELD_TS"/>
                        <constant>
                            <xsl:copy-of select="/symod:query/symod:parsedTime/@resolution"/>
                            <xsl:choose>
                                <xsl:when test="@_parsedSince"><xsl:value-of select="@_parsedSince"/></xsl:when>
                                <xsl:otherwise><xsl:value-of select="/symod:query/symod:parsedTime/@since"/></xsl:otherwise>
                            </xsl:choose>
                        </constant>
                    </constraint>
                </xsl:if>
                <xsl:if test="/symod:query/symod:parsedTime/@until | @_parsedUntil">
                    <constraint operator="&lt;">
                        <xsl:apply-templates select="." mode="FIELD_TS"/>
                        <constant>
                            <xsl:copy-of select="/symod:query/symod:parsedTime/@resolution"/>
                            <xsl:choose>
                                <xsl:when test="@_parsedUntil"><xsl:value-of select="@_parsedUntil"/></xsl:when>
                                <xsl:otherwise><xsl:value-of select="/symod:query/symod:parsedTime/@until"/></xsl:otherwise>
                            </xsl:choose>
                        </constant>
                    </constraint>
                </xsl:if>
            </xsl:for-each>
        </xsl:for-each>
        <xsl:if test="/symod:query/symod:parsedTime/@resolution">
            <!-- contrainte d'egalite entre : 1ere mesure de chaque entity + toutes les value_dates -->
            <xsl:variable name="first"
                          select="descendant::symod:entity/symod:value[(@_table='M' and not(preceding-sibling::symod:value/@_table='M')) or @_isDate='true'][not(@joinTime='false')][not(@_parsedSince|@_parsedUntil)][1]"/>
            <xsl:for-each select="descendant::symod:entity/symod:value[(@_table='M' and not(preceding-sibling::symod:value/@_table='M')) or @_isDate='true'][not(@joinTime='false')][not(@_parsedSince|@_parsedUntil)]">
                <xsl:if test="position() > 1">
                    <constraint operator="=">
                        <xsl:apply-templates select="$first" mode="FIELD_TS_ROUNDED"/>
                        <xsl:text> = </xsl:text>
                        <xsl:apply-templates select="." mode="FIELD_TS_ROUNDED"/>
                    </constraint>
                </xsl:if>
            </xsl:for-each>
        </xsl:if>
    </xsl:template>

    <xsl:template name="NULL_CONSTRAINTS">
        <xsl:for-each select="descendant::symod:entity">
            <xsl:for-each select="symod:value[not(@_isDate='true') or @joinTime='false'][not(symod:parsedConstraint)]">
                <xsl:choose>
                    <xsl:when test="@null='required'">
                        <constraint operator="IS">
                            <field name="{@TYPE}">
                                <xsl:attribute name="prefix"><xsl:apply-templates select="." mode="VALUE_ID"/></xsl:attribute>
                            </field>
                            <constant>NULL</constant>
                        </constraint>
                    </xsl:when>
                    <xsl:when test="@null='forbidden'">
                        <constraint operator="IS">
                            <field name="{@TYPE}">
                                <xsl:attribute name="prefix"><xsl:apply-templates select="." mode="VALUE_ID"/></xsl:attribute>
                            </field>
                            <constant>NOT NULL</constant>
                        </constraint>
                    </xsl:when>
                </xsl:choose>
            </xsl:for-each>
        </xsl:for-each>
    </xsl:template>

    <!-- ================================ SELECT ================================ -->

    <xsl:template match="symod:value/symod:print">
        <column>
            <xsl:for-each select="parent::symod:value">
                <xsl:attribute name="as">
                    <xsl:apply-templates select="." mode="NAME"/>
                </xsl:attribute>
            </xsl:for-each>
            <field>
                <xsl:for-each select="parent::symod:value">
                    <xsl:attribute name="name"><xsl:value-of select="@TYPE"/></xsl:attribute>
                    <xsl:attribute name="prefix"><xsl:apply-templates select="." mode="VALUE_ID"/></xsl:attribute>
                    <xsl:apply-templates select="symod:parsedFormula"/>
                </xsl:for-each>
            </field>
        </column>
    </xsl:template>
    <xsl:template match="symod:entity_value/symod:print">
        <column><xsl:copy-of select="@as"/>
            <_><xsl:apply-templates select="parent::symod:entity_value/symod:parsedFormula"/></_>
        </column>
    </xsl:template>
    <xsl:template match="symod:parsedFormula">
        <xsl:value-of select="@formula"/>
    </xsl:template>

    <!-- ================================ WHERE ================================ -->

    <xsl:template match="symod:and|symod:or">
        <xsl:copy><xsl:apply-templates/></xsl:copy>
    </xsl:template>

    <xsl:template match="symod:add|symod:substract|symod:multiply|symod:divide|symod:power">
        <xsl:copy><xsl:apply-templates/></xsl:copy>
    </xsl:template>

    <!-- see example attr_value_date_join_time -->
    <xsl:template match="symod:comparison">
        <xsl:variable name="aggregate" select="@aggregate"/>
        <constraint>
            <xsl:attribute name="operator">
                <xsl:choose>
                    <xsl:when test="@operator='like'">LIKE</xsl:when>
                    <xsl:when test="@operator='notLike'">NOT LIKE</xsl:when>
                    <xsl:when test="@operator='in'">IN</xsl:when>
                    <xsl:when test="@operator='notIn'">NOT IN</xsl:when>
                    <xsl:when test="@operator='equals'">=</xsl:when>
                    <xsl:when test="@operator='differs'">!=</xsl:when>
                    <xsl:when test="@operator='greaterThan'">&gt;</xsl:when>
                    <xsl:when test="@operator='greaterOrEquals'">&gt;=</xsl:when>
                    <xsl:when test="@operator='lowerThan'">&lt;</xsl:when>
                    <xsl:when test="@operator='lowerOrEquals'">&lt;=</xsl:when>
                    <xsl:when test="@operator='between'">BETWEEN</xsl:when>
                    <xsl:when test="@operator='notBetween'">NOT BETWEEN</xsl:when>
                </xsl:choose>
            </xsl:attribute>
            <xsl:for-each select="parent::symod:parsedConstraint/parent::symod:value">
                <field name="{@TYPE}">
                    <xsl:attribute name="prefix"><xsl:apply-templates select="." mode="VALUE_ID"/></xsl:attribute>
                    <xsl:if test="$aggregate"><xsl:value-of select="$aggregate"/>(self)</xsl:if>
                </field>
            </xsl:for-each>
            <xsl:apply-templates/>
        </constraint>
    </xsl:template>

    <xsl:template match="symod:comparison[@lowerBound]/symod:constant" priority="9">
        <constant><xsl:value-of select="parent::symod:comparison/@lowerBound"/> AND <xsl:value-of select="text()"/></constant>
    </xsl:template>
    <xsl:template match="symod:constant">
        <constant><xsl:value-of select="text()"/></constant>
    </xsl:template>
    <xsl:template match="symod:variable">
        <xsl:variable name="current" select="text()"/>
        <xsl:apply-templates select="ancestor::symod:query[parent::symod:query]//symod:*[symod:parsedConstraint/symod:declaration/@variable=$current]"/>
    </xsl:template>

    <xsl:template match="symod:entity/@constraint">
        <multi_constraint>(<xsl:call-template name="REPLACE">
            <xsl:with-param name="str" select="."/>
            <xsl:with-param name="replace">###</xsl:with-param>
            <xsl:with-param name="with">
                <xsl:apply-templates select="parent::symod:entity" mode="ENTITY_ID"/>
            </xsl:with-param>
        </xsl:call-template>)</multi_constraint>
    </xsl:template>
    <xsl:template name="REPLACE">
        <xsl:param name="str"/>
        <xsl:param name="replace"/>
        <xsl:param name="with"/>
        <xsl:choose>
            <xsl:when test="contains($str, $replace)">
                <xsl:value-of select="substring-before($str, $replace)"/>
                <xsl:value-of select="$with"/>
                <xsl:call-template name="REPLACE">
                    <xsl:with-param name="str" select="substring-after($str, $replace)"/>
                    <xsl:with-param name="replace" select="$replace"/>
                    <xsl:with-param name="with" select="$with"/>
                </xsl:call-template>
            </xsl:when>
            <xsl:otherwise>
                <xsl:value-of select="$str"/>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>

    <!-- ================================ COMMON ================================ -->

    <xsl:template match="symod:parsedTime">
        <xsl:choose>
            <xsl:when test="/symod:query/@resolution">
                <time/>
            </xsl:when>
            <xsl:otherwise>
                <!-- print time with no resolution (all measures) -->
                <xsl:apply-templates select="/symod:query/symod:query/descendant::symod:entity/symod:value[@_table='M'][1]">
                    <xsl:with-param name="timestamp">true</xsl:with-param>
                </xsl:apply-templates>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>
    <xsl:template match="symod:value">
        <xsl:param name="timestamp"/>
        <field value="{@TYPE}">
            <xsl:attribute name="prefix">
                <xsl:choose>
                    <xsl:when test="@TYPE='Id' and parent::symod:entity/symod:value/@_table">
                        <xsl:value-of select="parent::symod:entity/symod:value/@_table"/>
                    </xsl:when>
                    <xsl:when test="@TYPE='Id'">A</xsl:when>
                    <xsl:when test="@_table"><xsl:value-of select="@_table"/></xsl:when>
                    <xsl:otherwise>ERROR(1)</xsl:otherwise>
                </xsl:choose>
            </xsl:attribute>
            <xsl:attribute name="entity"><xsl:apply-templates select="parent::symod:entity" mode="ENTITY_ID"/></xsl:attribute>
            <xsl:copy-of select="symod:parsedConstraint/symod:*/@aggregate"/>
            <xsl:copy-of select="@name | @as"/>
            <xsl:if test="$timestamp">
                <xsl:variable name="ts">ts_<xsl:apply-templates select="parent::symod:entity" mode="ENTITY_ID"/></xsl:variable>
                <xsl:attribute name="timestamp">true</xsl:attribute>
                <xsl:attribute name="name"><xsl:value-of select="$ts"/>_<xsl:value-of select="@TYPE"/></xsl:attribute>
                <xsl:attribute name="as"><xsl:value-of select="$ts"/>_<xsl:value-of select="@TYPE"/></xsl:attribute>
            </xsl:if>
            <xsl:copy-of select="@*[starts-with(local-name(), '_')]"/>
        </field>
    </xsl:template>

    <xsl:template match="symod:value | symod:entity_value" mode="NAME">
        <xsl:choose>
            <xsl:when test="@name"><xsl:value-of select="@name"/></xsl:when>
            <xsl:otherwise><xsl:apply-templates select="." mode="VALUE_ID"/>_<xsl:value-of select="@_shortTYPE"/></xsl:otherwise>
        </xsl:choose>
    </xsl:template>

    <xsl:template match="symod:parsedTime" mode="FIELD_TS">
        <field prefix="ref">
            <xsl:attribute name="name">
                <xsl:choose>
                    <xsl:when test="@name"><xsl:value-of select="@name"/></xsl:when>
                    <xsl:otherwise>TS_REF</xsl:otherwise>
                </xsl:choose>
            </xsl:attribute>
        </field>
    </xsl:template>
    <xsl:template match="symod:value" mode="FIELD_TS">
        <field>
            <xsl:attribute name="name">
                <xsl:choose>
                    <xsl:when test="@_table='M'">TIMESTAMP</xsl:when>
                    <xsl:otherwise><xsl:value-of select="@TYPE"/></xsl:otherwise>
                </xsl:choose>
            </xsl:attribute>
            <xsl:attribute name="prefix"><xsl:apply-templates select="." mode="VALUE_ID"/></xsl:attribute>
        </field>
    </xsl:template>
    <xsl:template match="symod:value" mode="FIELD_TS_ROUNDED">
        <field>
            <xsl:attribute name="name">
                <xsl:choose>
                    <xsl:when test="@_table='M'">TIMESTAMP</xsl:when>
                    <xsl:otherwise><xsl:value-of select="@TYPE"/></xsl:otherwise>
                </xsl:choose>
            </xsl:attribute>
            <xsl:attribute name="prefix"><xsl:apply-templates select="." mode="VALUE_ID"/></xsl:attribute>
            <xsl:copy-of select="/symod:query/symod:parsedTime/@resolution"/>
        </field>
    </xsl:template>

    <xsl:template match="symod:entity" mode="JOIN_ENTITY">
        <from>
            <xsl:choose>
                <xsl:when test="symod:value[@_table='M'] and symod:value[@_table='A']">
                    <xsl:variable name="prefix1">
                        <xsl:choose>
                            <xsl:when test="symod:value[@join='true']/@_table">
                                <xsl:value-of select="symod:value[@join='true']/@_table"/>
                            </xsl:when>
                            <xsl:otherwise>M</xsl:otherwise>
                        </xsl:choose>
                    </xsl:variable>
                    <xsl:attribute name="table"><xsl:value-of select="$prefix1"/>_<xsl:value-of select="@TYPE"/></xsl:attribute>
                    <xsl:attribute name="as"><xsl:value-of select="$prefix1"/>_<xsl:apply-templates select="." mode="ENTITY_ID"/></xsl:attribute>
                    <xsl:apply-templates select="symod:value[symod:entity][@_table=$prefix1]" mode="JOIN_ENTITY"/>
                    <join parentField="Id" childField="Id">
                        <from>
                            <xsl:variable name="prefix2" select="translate($prefix1,'AM','MA')"/>
                            <xsl:attribute name="table"><xsl:value-of select="$prefix2"/>_<xsl:value-of select="@TYPE"/></xsl:attribute>
                            <xsl:attribute name="as"><xsl:value-of select="$prefix2"/>_<xsl:apply-templates select="." mode="ENTITY_ID"/></xsl:attribute>
                            <xsl:apply-templates select="symod:value[symod:entity][@_table=$prefix2 or @TYPE='Id']" mode="JOIN_ENTITY"/>
                        </from>
                    </join>
                </xsl:when>
                <xsl:when test="symod:value[@_table='M']">
                    <xsl:attribute name="table">M_<xsl:value-of select="@TYPE"/></xsl:attribute>
                    <xsl:attribute name="as">M_<xsl:apply-templates select="." mode="ENTITY_ID"/></xsl:attribute>
                    <xsl:apply-templates select="symod:value[symod:entity]" mode="JOIN_ENTITY"/>
                </xsl:when>
                <xsl:when test="symod:value[@_table='A' or @TYPE='Id']">
                    <xsl:attribute name="table">A_<xsl:value-of select="@TYPE"/></xsl:attribute>
                    <xsl:attribute name="as">A_<xsl:apply-templates select="." mode="ENTITY_ID"/></xsl:attribute>
                    <xsl:apply-templates select="symod:value[symod:entity]" mode="JOIN_ENTITY"/>
                </xsl:when>
            </xsl:choose>
        </from>
    </xsl:template>
    <xsl:template match="symod:value[symod:entity]" mode="JOIN_ENTITY">
        <xsl:for-each select="symod:entity">
            <join parentField="{parent::symod:value/@TYPE}" childField="{
                symod:value[@join='true' or (not(../symod:value/@join) and @link=../symod:value/@link)]/@TYPE}">
                <xsl:apply-templates select="." mode="JOIN_ENTITY"/>
            </join>
        </xsl:for-each>
    </xsl:template>

    <xsl:template match="symod:value" mode="VALUE_ID">
        <xsl:choose>
            <xsl:when test="self::symod:parsedTime">Time</xsl:when>
            <xsl:otherwise>
                <xsl:choose>
                    <xsl:when test="@TYPE='Id' and parent::symod:entity/symod:value/@_table">
                        <xsl:value-of select="parent::symod:entity/symod:value/@_table"/>
                    </xsl:when>
                    <xsl:when test="@TYPE='Id'">A</xsl:when>
                    <xsl:when test="@_table"><xsl:value-of select="@_table"/></xsl:when>
                    <xsl:otherwise>ERROR(4)</xsl:otherwise>
                </xsl:choose>
                <xsl:text>_</xsl:text>
                <xsl:apply-templates select="parent::symod:entity" mode="ENTITY_ID"/>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>

    <xsl:template match="symod:entity" mode="ENTITY_ID">
        <xsl:variable name="TYPE" select="@TYPE"/>
        <xsl:value-of select="@TYPE"/>
        <xsl:value-of select="count(preceding::symod:entity[@TYPE=$TYPE] | ancestor::symod:entity[@TYPE=$TYPE])"/>
    </xsl:template>
</xsl:stylesheet>
